
#line 1 "lc.rl"
/* lc -- Line encode a bitstream with output in WaveDrom format.
   Copyright (C) 2022 Remington Furman
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <string.h>
#include <getopt.h>
#include <stdlib.h>

struct line_code
{
    int cs;   /* Current state. */
    int bit;  /* Current bit value. */
};


#line 58 "lc.rl"



#line 34 "lc.c"
static const int line_code_start = 1;
static const int line_code_first_final = 14;
static const int line_code_error = 0;

static const int line_code_en_nrz_l = 15;
static const int line_code_en_nrz_i = 16;
static const int line_code_en_nrz_s = 17;
static const int line_code_en_nrz_m = 18;
static const int line_code_en_bph_l = 19;
static const int line_code_en_bph_s = 20;
static const int line_code_en_bph_m = 21;
static const int line_code_en_pass = 11;
static const int line_code_en_invert = 12;
static const int line_code_en_manch = 13;
static const int line_code_en_main = 1;


#line 61 "lc.rl"

void line_code_init(struct line_code *fsm, int first_bit, int start)
{
    fsm->bit = first_bit;

    
#line 59 "lc.c"
	{
	 fsm->cs = line_code_start;
	}

#line 67 "lc.rl"

    fsm->cs = start;
}

void line_code_execute(struct line_code *fsm, const char *data, int len)
{
    const char *p = data;
    const char *pe = data + len;

    
#line 75 "lc.c"
	{
	if ( p == pe )
		goto _test_eof;
	switch (  fsm->cs )
	{
case 1:
	goto tr0;
tr0:
#line 57 "lc.rl"
	{ p--; {goto st15;} }
	goto st14;
st14:
	if ( ++p == pe )
		goto _test_eof14;
case 14:
#line 91 "lc.c"
	goto st0;
st0:
 fsm->cs = 0;
	goto _out;
case 11:
	goto tr27;
tr27:
#line 54 "lc.rl"
	{ p--; {goto st15;} }
	goto st22;
st22:
	if ( ++p == pe )
		goto _test_eof22;
case 22:
#line 106 "lc.c"
	goto st0;
case 12:
	goto tr28;
tr28:
#line 55 "lc.rl"
	{ p--; {goto st16;} }
	goto st23;
st23:
	if ( ++p == pe )
		goto _test_eof23;
case 23:
#line 118 "lc.c"
	goto st0;
case 13:
	goto tr29;
tr29:
#line 56 "lc.rl"
	{ p--; {goto st19;} }
	goto st24;
st24:
	if ( ++p == pe )
		goto _test_eof24;
case 24:
#line 130 "lc.c"
	goto st0;
tr1:
#line 36 "lc.rl"
	{                 putchar('\n'); }
	goto st15;
st15:
	if ( ++p == pe )
		goto _test_eof15;
case 15:
#line 140 "lc.c"
	switch( (*p) ) {
		case 0: goto tr1;
		case 48: goto tr5;
		case 49: goto tr4;
	}
	goto st0;
tr3:
#line 35 "lc.rl"
	{                 putchar('.'); }
	goto st2;
tr5:
#line 30 "lc.rl"
	{ fsm->bit  = 0;  putchar('l'); }
	goto st2;
st2:
	if ( ++p == pe )
		goto _test_eof2;
case 2:
#line 159 "lc.c"
	switch( (*p) ) {
		case 0: goto tr1;
		case 48: goto tr3;
		case 49: goto tr4;
	}
	goto st0;
tr6:
#line 35 "lc.rl"
	{                 putchar('.'); }
	goto st3;
tr4:
#line 31 "lc.rl"
	{ fsm->bit  = 1;  putchar('h'); }
	goto st3;
st3:
	if ( ++p == pe )
		goto _test_eof3;
case 3:
#line 178 "lc.c"
	switch( (*p) ) {
		case 0: goto tr1;
		case 48: goto tr5;
		case 49: goto tr6;
	}
	goto st0;
tr7:
#line 36 "lc.rl"
	{                 putchar('\n'); }
	goto st16;
st16:
	if ( ++p == pe )
		goto _test_eof16;
case 16:
#line 193 "lc.c"
	switch( (*p) ) {
		case 0: goto tr7;
		case 48: goto tr10;
		case 49: goto tr9;
	}
	goto st0;
tr8:
#line 35 "lc.rl"
	{                 putchar('.'); }
	goto st4;
tr10:
#line 31 "lc.rl"
	{ fsm->bit  = 1;  putchar('h'); }
	goto st4;
st4:
	if ( ++p == pe )
		goto _test_eof4;
case 4:
#line 212 "lc.c"
	switch( (*p) ) {
		case 0: goto tr7;
		case 48: goto tr8;
		case 49: goto tr9;
	}
	goto st0;
tr11:
#line 35 "lc.rl"
	{                 putchar('.'); }
	goto st5;
tr9:
#line 30 "lc.rl"
	{ fsm->bit  = 0;  putchar('l'); }
	goto st5;
st5:
	if ( ++p == pe )
		goto _test_eof5;
case 5:
#line 231 "lc.c"
	switch( (*p) ) {
		case 0: goto tr7;
		case 48: goto tr10;
		case 49: goto tr11;
	}
	goto st0;
tr12:
#line 36 "lc.rl"
	{                 putchar('\n'); }
	goto st17;
st17:
	if ( ++p == pe )
		goto _test_eof17;
case 17:
#line 246 "lc.c"
	switch( (*p) ) {
		case 0: goto tr12;
		case 48: goto tr13;
		case 49: goto tr14;
	}
	goto st0;
tr14:
#line 35 "lc.rl"
	{                 putchar('.'); }
	goto st6;
tr13:
#line 32 "lc.rl"
	{ fsm->bit ^= 1;  putchar(fsm->bit ? 'h' : 'l'); }
	goto st6;
st6:
	if ( ++p == pe )
		goto _test_eof6;
case 6:
#line 265 "lc.c"
	switch( (*p) ) {
		case 0: goto tr12;
		case 48: goto tr13;
		case 49: goto tr14;
	}
	goto st0;
tr15:
#line 36 "lc.rl"
	{                 putchar('\n'); }
	goto st18;
st18:
	if ( ++p == pe )
		goto _test_eof18;
case 18:
#line 280 "lc.c"
	switch( (*p) ) {
		case 0: goto tr15;
		case 48: goto tr16;
		case 49: goto tr17;
	}
	goto st0;
tr16:
#line 35 "lc.rl"
	{                 putchar('.'); }
	goto st7;
tr17:
#line 32 "lc.rl"
	{ fsm->bit ^= 1;  putchar(fsm->bit ? 'h' : 'l'); }
	goto st7;
st7:
	if ( ++p == pe )
		goto _test_eof7;
case 7:
#line 299 "lc.c"
	switch( (*p) ) {
		case 0: goto tr15;
		case 48: goto tr16;
		case 49: goto tr17;
	}
	goto st0;
tr18:
#line 36 "lc.rl"
	{                 putchar('\n'); }
	goto st19;
st19:
	if ( ++p == pe )
		goto _test_eof19;
case 19:
#line 314 "lc.c"
	switch( (*p) ) {
		case 0: goto tr18;
		case 48: goto tr19;
		case 49: goto tr20;
	}
	goto st0;
tr19:
#line 30 "lc.rl"
	{ fsm->bit  = 0;  putchar('l'); }
#line 31 "lc.rl"
	{ fsm->bit  = 1;  putchar('h'); }
	goto st8;
tr20:
#line 31 "lc.rl"
	{ fsm->bit  = 1;  putchar('h'); }
#line 30 "lc.rl"
	{ fsm->bit  = 0;  putchar('l'); }
	goto st8;
st8:
	if ( ++p == pe )
		goto _test_eof8;
case 8:
#line 337 "lc.c"
	switch( (*p) ) {
		case 0: goto tr18;
		case 48: goto tr19;
		case 49: goto tr20;
	}
	goto st0;
tr21:
#line 36 "lc.rl"
	{                 putchar('\n'); }
	goto st20;
st20:
	if ( ++p == pe )
		goto _test_eof20;
case 20:
#line 352 "lc.c"
	switch( (*p) ) {
		case 0: goto tr21;
		case 48: goto tr22;
		case 49: goto tr23;
	}
	goto st0;
tr22:
#line 33 "lc.rl"
	{ fsm->bit ^= 1;  putchar(fsm->bit ? 'h' : 'l');
                     fsm->bit ^= 1;  putchar(fsm->bit ? 'h' : 'l'); }
	goto st9;
tr23:
#line 32 "lc.rl"
	{ fsm->bit ^= 1;  putchar(fsm->bit ? 'h' : 'l'); }
#line 35 "lc.rl"
	{                 putchar('.'); }
	goto st9;
st9:
	if ( ++p == pe )
		goto _test_eof9;
case 9:
#line 374 "lc.c"
	switch( (*p) ) {
		case 0: goto tr21;
		case 48: goto tr22;
		case 49: goto tr23;
	}
	goto st0;
tr24:
#line 36 "lc.rl"
	{                 putchar('\n'); }
	goto st21;
st21:
	if ( ++p == pe )
		goto _test_eof21;
case 21:
#line 389 "lc.c"
	switch( (*p) ) {
		case 0: goto tr24;
		case 48: goto tr25;
		case 49: goto tr26;
	}
	goto st0;
tr26:
#line 33 "lc.rl"
	{ fsm->bit ^= 1;  putchar(fsm->bit ? 'h' : 'l');
                     fsm->bit ^= 1;  putchar(fsm->bit ? 'h' : 'l'); }
	goto st10;
tr25:
#line 32 "lc.rl"
	{ fsm->bit ^= 1;  putchar(fsm->bit ? 'h' : 'l'); }
#line 35 "lc.rl"
	{                 putchar('.'); }
	goto st10;
st10:
	if ( ++p == pe )
		goto _test_eof10;
case 10:
#line 411 "lc.c"
	switch( (*p) ) {
		case 0: goto tr24;
		case 48: goto tr25;
		case 49: goto tr26;
	}
	goto st0;
	}
	_test_eof14:  fsm->cs = 14; goto _test_eof; 
	_test_eof22:  fsm->cs = 22; goto _test_eof; 
	_test_eof23:  fsm->cs = 23; goto _test_eof; 
	_test_eof24:  fsm->cs = 24; goto _test_eof; 
	_test_eof15:  fsm->cs = 15; goto _test_eof; 
	_test_eof2:  fsm->cs = 2; goto _test_eof; 
	_test_eof3:  fsm->cs = 3; goto _test_eof; 
	_test_eof16:  fsm->cs = 16; goto _test_eof; 
	_test_eof4:  fsm->cs = 4; goto _test_eof; 
	_test_eof5:  fsm->cs = 5; goto _test_eof; 
	_test_eof17:  fsm->cs = 17; goto _test_eof; 
	_test_eof6:  fsm->cs = 6; goto _test_eof; 
	_test_eof18:  fsm->cs = 18; goto _test_eof; 
	_test_eof7:  fsm->cs = 7; goto _test_eof; 
	_test_eof19:  fsm->cs = 19; goto _test_eof; 
	_test_eof8:  fsm->cs = 8; goto _test_eof; 
	_test_eof20:  fsm->cs = 20; goto _test_eof; 
	_test_eof9:  fsm->cs = 9; goto _test_eof; 
	_test_eof21:  fsm->cs = 21; goto _test_eof; 
	_test_eof10:  fsm->cs = 10; goto _test_eof; 

	_test_eof: {}
	_out: {}
	}

#line 77 "lc.rl"
}

int line_code_finish(struct line_code *fsm)
{
    if (fsm->cs == line_code_error) {
        return -1;
    }

    if (fsm->cs >= line_code_first_final) {
        return 1;
    }
    return 0;
}

#define NUM_ELEMENTS(a) (sizeof(a)/sizeof(a[0]))
#define XSTRINGIFY(s) #s
#define STRINGIFY(s) XSTRINGIFY(s)
#define MACHINE(name, desc) { STRINGIFY(name), desc, &line_code_en_ ## name }

struct { const char *name; char* desc; const int *start; } machines[] = {
    MACHINE(nrz_l,  "Non-return-to-zero level"),
    MACHINE(nrz_i,  "Non-return-to-zero invert"),
    MACHINE(nrz_s,  "Non-return-to-zero space"),
    MACHINE(nrz_m,  "Non-return-to-zero mark"),
    MACHINE(bph_l,  "Bi-phase level"),
    MACHINE(bph_s,  "Bi-phase space"),
    MACHINE(bph_m,  "Bi-phase mark"),
    MACHINE(pass,   "Passthrough (alias of nrz_l)"),
    MACHINE(invert, "Invert (alias of nrz_i)"),
    MACHINE(manch,  "Manchester (alias of bph_l)"),
};

void usage(char *name) {
    printf("%s [flags] input" , name);
    printf("    input\n"
           "        One or more strings of 0's and 1's\n"
           "    -i n\n"
           "        Initial line state (default: 0)\n"
           "    -c name\n"
           "        Line code to use (default: pass)\n");
    printf("\nAvailable codes:\n");
    for (int i = 0; i < NUM_ELEMENTS(machines); i++) {
        printf("  %-7s- %s\n", machines[i].name, machines[i].desc);
    }
}

int main(int argc, char **argv)
{
    struct line_code lc;

    int first_bit = 0;
    int start = line_code_en_nrz_l;

    int opt;
    while ((opt = getopt(argc, argv, "f:m:h")) != -1) {
        switch (opt) {
            case 'f': {
                first_bit = !!atoi(optarg);
                break;
            }
            case 'm': {
                int found = 0;
                for (int i = 0; i < NUM_ELEMENTS(machines); i++) {
                    if (strcmp(optarg, machines[i].name) == 0) {
                        start = *machines[i].start;
                        found = 1;
                        break;
                    }
                }
                if (!found) {
                    printf("Invalid machine name: %s\n", optarg);
                    usage(argv[0]);
                    exit(EXIT_FAILURE);
                }
                break;
            }
            case 'h': {
                usage(argv[0]);
                exit(EXIT_SUCCESS);
                break;
            }
            default: {
                usage(argv[0]);
                exit(EXIT_FAILURE);
                break;
            }
        }
    }

    /* Initialize state machine */
    line_code_init(&lc, first_bit, start);

    for (int i = optind; i < argc; i++) {
        /* Encode each input argument. */
        line_code_execute(&lc, argv[i], strlen(argv[i])+1);
    }

    if (line_code_finish(&lc) != 1) {
        fprintf(stderr, "%s: bad input\n", argv[0]);
        usage(argv[0]);
        exit(EXIT_FAILURE);
    }

    return 0;
}
